
import matplotlib.pyplot as plt
import matplotlib.colors as colors
import numpy as np
import math


''' Aufgabe 3 '''



def convolve(image, matrix):
           
        # Neues Bild anlegen:
        h, w = image.shape
        result = np.zeros((h, w))        
        
        # Alle Pixel durchgehen:
        for y in xrange(h):
            print "Convolution working at row", y, "..."
            for x in xrange(w):
                result[y, x] = convolveAt(image, x, y, matrix)
        
        # Ergebnis zurueckgeben:
        return result
    
    
def convolveAt(image, x, y, matrix):
        h, w = image.shape
        m, n = matrix.shape
        m = int(math.floor(m/2))
        n = int(math.floor(n/2))
        sum = 0
        for my in xrange(-m, m+1):
            for mx in xrange(-n, n+1):
                sum += matrix[my+m, mx+n] * image[max(0, min(h-1, y + my)), max(0, min(w-1, x + mx))]
        return sum
    


def createGaussianKernel(n, sigma):
    
    a = 1.0/(sigma * math.sqrt(2*math.pi) )
    b = 1.0/(2*sigma*sigma)
    
    matrix = np.ones((n, n))
    for y in xrange(n):
        for x in xrange(n):
            xV = (x - int(n/2.0))
            yV = (y - int(n/2.0))
            matrix[y, x] = a * math.exp(-b * (xV*xV+yV*yV) )
    
    matrix = matrix / np.sum(matrix) #Normalisieren.
    return matrix


def gaussianBlur(image, n, sigma):
    matrix = createGaussianKernel(n, sigma)
    return convolve(image, matrix)
    
    



def medianFilter(image, m=1, n=1):
    
    # Neues Bild anlegen:
    h, w = image.shape
    result = np.zeros((h, w))        
    
    # Alle Pixel durchgehen:
    for y in xrange(h):
        for x in xrange(w):
            result[y, x] = medianFilterAt(image, x, y, m, n)[1]
    
    # Ergebnis zurueckgeben:
    return result



def medianFilterAt(image, x, y, m, n):
    
    h, w = image.shape
        
    valueList = [] #Hier werden die Pixelwerte gespeichert.
    
    # Alle Pixel im aktuellen "Fenster" durchgehen:  
    for i in xrange(max(0, y-m), min(h, y+m+1)):
        for j in xrange(max(0, x-n), min(w, x+n+1)):            
            valueList.append(image[i, j])
    
    valueList = sorted(valueList) #Werte aufsteigend sortieren.
    middle = len(valueList)/2 #Mittlerer Wert.
    minV = valueList[0]
    maxV = valueList[len(valueList)-1]
    even = (m*n) % 2 == 0 #Gerade oder ungerade Anzahl an Werten?
    medV = 0
    if(even): #Bei gerader Anzahl den Durchschnitt der mitteleren beiden Werte zurueckgeben.
        medV = (valueList[int(math.floor(middle))] + valueList[int(math.ceil(middle))])/2.0
    else:
        medV = valueList[middle] #Bei ungerader Anzahl, den mittleren Wert zurueckgeben.
    return minV, medV, maxV, image[y, x]
    
        
          
      
              
def aufgabe3():
    
    saltyLena = plt.imread('lena-saltpepper.png')
    cleanLena = medianFilter(saltyLena, 1, 1) #3x3 Median
    input = plt.imread('nerds_noisy.png')
    resultGauss = gaussianBlur(input, 3, 1.5) #3x3 Gauss
    resultMedian = medianFilter(input, 1, 1) #3x3 Median
    plt.figure(1)
    plt.subplot(231)
    plt.imshow(saltyLena, cmap=plt.cm.gray,norm=colors.NoNorm())
    plt.subplot(232)
    plt.imshow(cleanLena, cmap=plt.cm.gray,norm=colors.NoNorm())
    plt.subplot(234)
    plt.imshow(input, cmap=plt.cm.gray,norm=colors.NoNorm())
    plt.subplot(235)
    plt.imshow(resultGauss, cmap=plt.cm.gray, norm=colors.NoNorm())
    plt.subplot(236)
    plt.imshow(resultMedian, cmap=plt.cm.gray, norm=colors.NoNorm())
    plt.show()  
    

    

aufgabe3()       


'''Vergleich der Verfahren:
Zur Entfernung von Salt-and-Pepper-Noise eignet sich der Medianfilter viel besser.
Der Gaussfilter verwischt das Rauschen ein wenig, aber es ist dennoch nach wie vor
sichtbar, z.B. im unteren Teil der Schachtel.
Beim Medianfilter wird der eigentliche Bildinhalt auch nicht so stark veraendert;
dies erkennt man z.B. gut an der kleinen Schrift auf der Schachtel 
und natuerlich an Lena.
'''

