'''
@author: svenkratz
'''
from Tkinter import *

import matplotlib
matplotlib.use('TkAgg')

import matplotlib.pyplot as plt
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg, NavigationToolbar2TkAgg
from matplotlib.figure import Figure
import numpy as np

import math


class TKMatplotlibImage:
	'''Diese Klasse bettet ein Matplotlib-Plot in einem TkInter-Widget ein'''
	def __init__(self, master, image):
		self.master = master
		self.image = image
		self.figure = Figure(figsize=(4,5), dpi=100)
		self.plot =  self.figure.add_subplot(111)
		plt.gray()
		self.plot.imshow(self.image)
		self.canvas = FigureCanvasTkAgg(self.figure, master=master)
		self.canvas.show()
		self.canvas.get_tk_widget().pack()
	
	def replot(self, normalize = False):
		'''Plot neu zeichnen'''
		print "Replot! Norm=",normalize
		self.plot.clear()
		if normalize:
			self.plot.imshow(self.image)
		else:
			self.plot.imshow(self.image, norm = matplotlib.colors.NoNorm())
		self.canvas.show()
		self.canvas.get_tk_widget().pack()




class ConvolveDialog:
	'''Zeigt einen Dialog an, ueber den eine Kernel-Matrix eingegeben werden kann'''
	def __init__(self, master, input_window, output_window, k_size=3):
		
		self.entries = []
		# Textfelder in Matrixform verteilen
		for i in range(k_size):
			row = []
			for j in range(k_size):
				print "Entry",i,j
				entry = Entry(master=master, width=4)
				if i == k_size / 2 and j == k_size / 2:
					entry.insert(0,1.0)
				else:
					entry.insert(0,0.0)
				entry.grid(row=i,column=j)
				row.append(entry)
				
			self.entries.append(row)
		
		
		self.norm_checked = IntVar(0)
		self.norm_checkButton = Checkbutton(master, text="normalize output", variable = self.norm_checked)
		self.norm_checkButton.select()
		self.norm_checkButton.grid(row=k_size, columnspan=k_size)
		
		# Button, mit dem die Konvolution aufgerufen wird, Ruft bei Click self.command auf
		self.button = Button(master, text="Convolve", command=self.command)
		self.button.grid(row=k_size+1, columnspan=k_size)
		
		self.input_window = input_window
		self.output_window = output_window
		
		self.conv_matrix= None
		
		self.getConvMatrix()
		print "Matrix", self.conv_matrix
		self.command()
		

		
	def command(self):
		##### Hier kann was mit dem Ausgabebild gemacht werden
		
		# holt die aktuelle Matrix des Kernes
		self.getConvMatrix()
		print "Starting convolution"
		# Konvolution des Eingabebildes
		convolved_image = self.convolve(self.input_window.image, self.conv_matrix)
		self.output_window.image = convolved_image
		
		####
		# Neuzeichnen des Ausgabeplots
		###
		normalize = self.norm_checked.get()
		output_window.replot(normalize)

			
	

	###### remove for uebung!!!
	
	def getConvMatrix(self):
		#matrix = np.zeros((k_size,k_size))
		entries = [[float(e.get()) for e in row] for row in self.entries]
		# transponieren, da Eintraege 90 Grad gedreht sind
		self.conv_matrix = np.array(entries).transpose()
		print self.conv_matrix
	
		
	def convolve(self, image, matrix):
		#####################################
		#
		#  Hier Konvolution Implementieren!! 
		#
		#####################################
		
		self.test = 0
		
		# Neues Bild anlegen:
		h, w = image.shape
		result = np.zeros((h, w))		
		
		# Alle Pixel durchgehen:
		for y in xrange(h):
			for x in xrange(w):
				result[y, x] = self.convolveAt(image, x, y, matrix)
		
		# Ergebnis zurueckgeben:
		print 'AUFRUFE:', self.test
		return result
	
	
	def convolveAt(self, image, x, y, matrix):
		h, w = image.shape
		m, n = matrix.shape
		m = int(math.floor(m/2))
		n = int(math.floor(n/2))
		sum = 0
		for my in xrange(-m, m+1):
			for mx in xrange(-n, n+1):
				self.test += 1
				# eigentlich:matrix[my+m, mx+n], aber da die matrix oben transponiert wurde...
				sum += matrix[mx+n, my+m] * image[max(0, min(h-1, y + my)), max(0, min(w-1, x + mx))]
		return sum
	
	
	
	def convolveSeparated(self, image, matrix):
		
		self.test = 0
				
		# Neues Bild anlegen:
		h, w = image.shape
		result = np.zeros((h, w))		
		
		
		# Alle Pixel durchgehen (Konvolution nur entlang Spalten):
		for y in xrange(h):
			for x in xrange(w):
				#Hier nur aufsummieren (matrix = ones), noch keine Beachtung der Gewichte (da sonst doppelt,
				#z.B. 3x3 Filter: 1/9 -> 1/9 entlang X mal 1/9 entlang Y => 1/81 der ursprgl. Werte).
				#Durch die Normalisierung des Bildes bei der Anzeige wuerde man hier in Python allerdings
				#keinen Unterschied sehen.
				result[y, x] = self.convolveAt1DCols(image, x, y, np.ones_like(matrix[0])) 
		# Alle Pixel durchgehen (Konvolution nur entlang Zeilen):
		for y in xrange(h):
			for x in xrange(w):
				result[y, x] = self.convolveAt1DRows(result, x, y, matrix[0])
		
		# Ergebnis zurueckgeben:
		print 'AUFRUFE:', self.test
		return result
	
	# Betrachtet Pixel ueber und unter dem Zentrum:
	def convolveAt1DCols(self, image, x, y, matrix):
		h = image.shape[0]
		m = matrix.shape[0]
		m = int(math.floor(m/2))
		sum = 0
		for mi in xrange(-m, m+1):
			self.test += 1
			sum += matrix[mi+m] * image[max(0, min(h-1, y + mi)), x]
		return sum
	
	# Betrachtet Pixel links und rechts neben dem Zentrum:
	def convolveAt1DRows(self, image, x, y, matrix):
		w = image.shape[1]
		n = matrix.shape[0]
		n = int(math.floor(n/2))
		sum = 0
		for mi in xrange(-n, n+1):
			self.test += 1
			sum += matrix[mi+n] * image[y, max(0, min(w-1, x + mi))]
		return sum
	
	
		
	''' Textaufgaben:
	1a 1) Verschiebung nach rechts unten um 1 Pixel.
	1a 2 Leichtes Blur entlang der x-Achse.
	1a 3) Leichtes Blur entlang der y-Achse.
	1a 4) Diagonales "Verwackeln" des Bildes. Erzeugt eine Art Geisterbild-Effekt.
	1a 5) Vertikale Kanten finden, von dunkel zu hell -> hell (1. Ableitung).
	1a 6) Vertikale Kanten finden, geglaettet in beide Richtungen (2. Ableitung). 
	
	
	1b) Methode(n) siehe oben: convolveSeparated
	
	Die Separierbarkeit ergibt sich direkt aus der Formel
		SummeY SummeX 1/filtergroesse * I_xy
	da sich der Faktor 1/filtergroesse beliebig ausklammern laesst.	
	D.h. man koennte z.B. die Werte entlang X aufsummieren, dann entlang Y 
	und mit 1/filtergroesse den Durchschnitt bilden.
	
	Meine Implementierung oben entspricht: SummeY 1/filtergroesse * SummeX I_xy
	Also ergibt sich die Separation konkret durch:
	1. tmp = SummeX I_xy
	2. SummeY 1/filtergroesse * tmp
		
	Zum Testen habe ich den Aufruf von convolve im Code der Angabe gegen convolveSeparated ersetzt.
	Anmerkung: Mein Algorithmus verwendet nur die Werte in der ersten Spalte der Matrix, da beim
	Boxcar-Filter ja ohnehin alle Eintraege der Matrix gleich sind.
	Das Attribut self.test zaehlt die Aufrufe von Pixeln und wird am Ende auf der Konsole ausgegeben.
	Ergebnis:
		Ohne Nutzung der Separierbarkeit sind es 607500 Aufrufe fuer das Testbild.
		Mit Nutzung der Separierbarkeit sind es hingegen nur 405000 Aufrufe.
	
	
	'''
	
	
	
# Root-Fenster
root = Tk()
ente = plt.imread('ente.png')
output = np.zeros(ente.shape)
# Neues Toplevel-Fenster fuer das Eingabebild
input_top = Toplevel(root)
input_top.title("Input Image")
# Neues Toplevel-Fenster fuer das Ausgabebild
output_top = Toplevel(root)
output_top.title("Output Image")
root.title("Convolution Matrix")


input_window = TKMatplotlibImage(input_top, ente)
output_window = TKMatplotlibImage(output_top, output)

# Dialog fuer die Eingabe einer Konvolutionsmatrix
convolve_dialog = ConvolveDialog(root, input_window, output_window)

# Geometry verschiebt die einzelnen Fenster
# muss ggf. je nach Bildschirmaufloesung angepasst werden
input_top.geometry('800x600+50+50')
output_top.geometry('800x600+950+50')
root.geometry('200x200+850+50')
root.deiconify()
root.mainloop()